from typing import Mapping, Any, Optional

import shapely
from shapely import Point
from shapely.geometry.base import BaseGeometry

from jord.shapely_utilities import dilate
from .morphology import closing, opening

__all__ = [
    "zero_buffer",
    "deflimmer",
    "clean_geometry",
    "unflimmer",
    "sanitise",
    "clean_shape",
]


def zero_buffer(
    geom: BaseGeometry,
) -> BaseGeometry:
    return dilate(geom, distance=0)


def clean_shape(
    shape: shapely.geometry.base.BaseGeometry,
) -> shapely.geometry.base.BaseGeometry:
    """
    removes self-intersections and duplicate points

    :param shape: The shape to cleaned
    :return: the cleaned shape
    """
    return zero_buffer(shape).simplify(0)


def deflimmer(geom: BaseGeometry, eps: float = 1e-7) -> BaseGeometry:
    """

    :param geom:
    :param eps:
    :return:
    """
    return opening(closing(geom, distance=eps), distance=eps)


clean_geometry = unflimmer = deflimmer


def sanitise(
    geom: BaseGeometry,
    *args: callable,
    kwargs: Optional[Mapping[callable, Mapping[str, Any]]] = None
) -> BaseGeometry:
    """
      #A positive distance produces a dilation, a negative distance an erosion. A very small or zero distance may sometimes be used to “tidy” a polygon.

    :param geom: The shape to sanitised
    :param args: The sanitisation callable steps
    :param kwargs: The sanitisation callable step kwargs in mappings with callable as key then sub-mapping is kwargs for callable
    :return: The sanitised shape
    """

    if kwargs is None:
        kwargs = {}

    if not len(args):
        args = (zero_buffer, deflimmer)

    for f in args:
        if f in kwargs:
            geom = f(geom, **(kwargs[f]))
        else:
            geom = f(geom)

    return geom


if __name__ == "__main__":

    def ausdhasu():
        p = Point((1, 1))
        print(clean_shape(p))

    ausdhasu()



#!/usr/bin/env python3
import base64
import json
import pickle
from enum import Enum
from typing import Tuple, Sequence, Any, Callable, Mapping, Union

__all__ = ["build_package", "read_package"]

from sorcery import assigned_names

from jord.qlive_utilities.procedures import QliveRPCMethodEnum, QliveRPCMethodMap


class SerialisationMethodEnum(Enum):
    json, pickle = assigned_names()


SERIALISATION_METHOD = SerialisationMethodEnum.pickle
VERBOSE = False


def build_package(
    method: QliveRPCMethodEnum, *args: Any, **kwargs
) -> Union[bytes, str]:
    """

    :param method:
    :param args:
    :return:
    """
    if VERBOSE:
        print(type(method), method, args)

    if not isinstance(method, QliveRPCMethodEnum):
        assert method
        assert isinstance(method, str), method
        method = QliveRPCMethodEnum(method)

    if VERBOSE:
        print(type(method.value), method.value, args)

    if SERIALISATION_METHOD == SerialisationMethodEnum.pickle:
        return pickle.dumps({"method": method.value, "args": args, "kwargs": kwargs})
    elif SERIALISATION_METHOD == SerialisationMethodEnum.json:
        return json.dumps(
            {"method": method.value, "args": args, "kwargs": kwargs}
        )  # TODO: ?
    elif False:
        ...
        # return base64.b64encode(str({"method": method.value, "args": args}).encode("ascii"))
    else:
        raise NotImplemented


def read_package(package: bytes) -> Tuple[Callable, Sequence[Any], Mapping[str, Any]]:
    """

    :param package:
    :return:
    """
    if SERIALISATION_METHOD == SerialisationMethodEnum.json:
        str_dict = (
            base64.b64decode(package)
            .decode("ascii")
            .replace(
                # Json library convert string dictionary to real dictionary type. Double quotes is standard format for json
                "'",
                '"',
            )
        )
        res_dict = json.loads(str_dict)  # convert string dictionary to dict format
    elif SERIALISATION_METHOD == SerialisationMethodEnum.pickle:
        res_dict = pickle.loads(package)
    else:
        raise NotImplemented
    return (
        QliveRPCMethodMap[QliveRPCMethodEnum(res_dict["method"])],
        res_dict["args"],
        res_dict["kwargs"],
    )


if __name__ == "__main__":
    print(read_package(build_package(method=QliveRPCMethodEnum.add_wkt)))
    print(read_package(build_package(method=QliveRPCMethodEnum.add_wkb)))
    print(read_package(build_package(method=QliveRPCMethodEnum.add_dataframe_layer)))



#!/usr/bin/env python3

__author__ = "heider"
__doc__ = r"""

           Created on 1/23/23
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()

from .degrees_of_freedom import *
from .tracing import *
from .center_line import *
from .intersections import *



#!/usr/bin/env python3

import random
from itertools import cycle
from typing import Callable, Generator, Iterable, Sized

# noinspection PyUnresolvedReferences
from qgis.PyQt.QtGui import QColor

# noinspection PyUnresolvedReferences
from qgis.core import (
    QgsCategorizedSymbolRenderer,
    QgsRendererCategory,
    QgsSymbol,
    QgsVectorLayer,
)

# noinspection PyUnresolvedReferences
from qgis.utils import iface
from warg import QuadNumber, TripleNumber, n_uint_mix_generator_builder

__all__ = ["categorise_layer"]


def random_rgb(mix: TripleNumber = (255, 255, 255)) -> TripleNumber:
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    return (red, green, blue)


def random_rgba(mix: QuadNumber = (1, 1, 1, 1)) -> QuadNumber:
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    alpha = random.randrange(0, mix[3])
    return (red, green, blue, alpha)


def random_color_generator() -> TripleNumber:
    while 1:
        yield random_rgb()


def random_color_alpha_generator() -> QuadNumber:
    while 1:
        yield random_rgba()


def categorise_layer(
    layer: QgsVectorLayer,
    field_name: str = "layer",
    iterable: Iterable = n_uint_mix_generator_builder(255, 255, 255, 255),
) -> None:
    """

    https://qgis.org/pyqgis/3.0/core/Vector/QgsVectorLayer.html
    https://qgis.org/pyqgis/3.0/core/other/QgsFields.html

    :param layer:
    :param field_name:
    :param iterable:
    :return:
    """

    if isinstance(iterable, Sized):
        # noinspection PyTypeChecker
        iterable = cycle(iterable)

    if isinstance(iterable, Callable) and not isinstance(iterable, Generator):
        # noinspection PyCallingNonCallable
        iterable = iterable()  # Compat

    color_iter = iter(iterable)

    available_field_names = layer.fields().names()

    assert (
        field_name in available_field_names
    ), f"Did not find {field_name=} in {available_field_names=}"

    render_categories = []
    for cat in layer.uniqueValues(layer.fields().indexFromName(field_name)):
        if cat:
            sym = QgsSymbol.defaultSymbol(layer.geometryType())
            col = next(color_iter)

            if len(col) == 3:
                col = (*col, 255)

            sym.setColor(QColor(*col))

            render_categories.append(
                QgsRendererCategory(cat, symbol=sym, label=str(cat), render=True)
            )

    if True:  # add default
        sym = QgsSymbol.defaultSymbol(layer.geometryType())
        col = next(color_iter)

        if len(col) == 3:
            col = (*col, 255)

        sym.setColor(QColor(*col))

        render_categories.append(
            QgsRendererCategory("", symbol=sym, label="default", render=True)
        )

        if False:
            # render_categories.append(QgsRendererCategory()) # crashes qgis
            render_categories.append(
                QgsRendererCategory([], symbol=sym, label="EmptyList", render=True)
            )
            render_categories.append(
                QgsRendererCategory("", symbol=sym, label="EmptyString", render=True)
            )
            render_categories.append(
                QgsRendererCategory("None", symbol=sym, label="None", render=True)
            )

    layer.setRenderer(QgsCategorizedSymbolRenderer(field_name, render_categories))
    layer.triggerRepaint()
    iface.layerTreeView().refreshLayerSymbology(layer.id())



#!/usr/bin/env python3


from enum import Enum

import shapely
from shapely.geometry import (
    Point,
    LineString,
    LinearRing,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
    GeometryCollection,
)

__all__ = ["ShapelyGeometryTypesEnum", "is_multi"]


class ShapelyGeometryTypesEnum(Enum):
    """
    This enum is useful for exhaustively iterating possible shapely types.
    """

    point = Point  # Point(*args) # A geometry type that represents a single coordinate with x,y and possibly z values.

    line_string = LineString  # LineString([coordinates]) # A geometry type composed of one or more line segments.

    linear_ring = LinearRing  # LinearRing([coordinates]) # A geometry type composed of one or more line segments that forms a closed loop.

    polygon = Polygon  # Polygon([shell, holes]) # A geometry type representing an area that is enclosed by a linear ring.

    multi_point = (
        MultiPoint  # MultiPoint([points]) # A collection of one or more Points.
    )

    multi_line_string = MultiLineString  # MultiLineString([lines]) # A collection of one or more LineStrings.

    multi_polygon = (
        MultiPolygon  # MultiPolygon([polygons]) # A collection of one or more Polygons.
    )

    geometry_collection = GeometryCollection  # GeometryCollection([geoms]) # A collection of one or more geometries that  may contain more than  one type of geometry.


MULTI_GEOMS = (
    shapely.MultiPolygon,
    shapely.MultiPoint,
    shapely.MultiLineString,
    shapely.GeometryCollection,
)


def is_multi(geom: shapely.geometry.base.BaseGeometry) -> bool:
    """
    Tests whether a geometry is multi-geometry

    :param geom: geometry to be tested
    :return: whether a geometry is multi-geometry
    """
    return isinstance(geom, MULTI_GEOMS)


if __name__ == "__main__":
    print([p.value.__name__ for p in ShapelyGeometryTypesEnum])



import uuid

__all__ = ["randomize_field", "randomize_sub_tree_field"]


def randomize_sub_tree_field(selected_nodes, field_name: str) -> None:
    # noinspection PyUnresolvedReferences
    from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer

    for group in selected_nodes:
        if isinstance(group, QgsLayerTreeLayer):
            randomize_field(group, field_name=field_name)
        elif isinstance(group, QgsLayerTreeGroup):
            randomize_sub_tree_field(group.children(), field_name=field_name)
        else:
            ...


def randomize_field(tree_layer, field_name: str) -> None:  #: QgsLayerTreeLayer
    """
        https://qgis.org/pyqgis/3.28/core/QgsVectorLayer.html#qgis.core.QgsVectorLayer.changeAttributeValues

    changeAttributeValues
    fid: int, newValues: Dict[int, Any], oldValues: Dict[int, Any] = {}, skipDefaultValues: bool = False

        :param field_name:
        :param tree_layer:
        :return:
    """
    layer = tree_layer.layer()

    field_idx = layer.fields().indexFromName(field_name)

    if field_idx >= 0:
        layer.startEditing()
        layer.beginEditCommand(f"Regenerate {field_name}")
        for i in range(layer.featureCount() + 1):
            layer.changeAttributeValue(i, field_idx, uuid.uuid4().hex)

        # layer.rollBack()
        layer.endEditCommand()
        layer.commitChanges()



#!/usr/bin/env python3

import statistics
from typing import Generator, Iterable, List, Sequence, Tuple, Union

import shapely
from shapely.geometry import (
    LineString,
    MultiLineString,
    MultiPolygon,
    Polygon,
)
from shapely.geometry.base import BaseGeometry
from warg import Number, pairs

from jord.shapely_utilities.base import sanitise
from .lines import segments
from .rings import ensure_ccw_ring, ensure_cw_ring

__all__ = [
    "explode_polygons",
    "polygon_has_interior_rings",
    "iter_polygons",
    "discard_holes",
    "get_coords_from_polygonal_shape",
    "get_polygonal_shape_from_coords",
]


def polygon_has_interior_rings(polygon: Polygon) -> bool:
    """

    :param polygon:
    :return:
    """
    return len(polygon.interiors) > 0


def mean_std_dev_line_length(geom: BaseGeometry) -> Tuple[float, float]:
    """

    :param geom:
    :return:
    """
    line_lengths = []
    if isinstance(geom, LineString):
        for segment in segments(geom):
            line_lengths.append(segment.length)
    elif isinstance(geom, MultiLineString):
        for li in geom.geoms:
            for segment in segments(li):
                line_lengths.append(segment.length)
    elif isinstance(geom, (Polygon, MultiPolygon)):
        exterior_rings, interior_rings = extract_poly_rings(geom)
        for ex_ring in exterior_rings:
            for segment in segments(ex_ring):
                line_lengths.append(segment.length)
        for in_ring in interior_rings:
            for segment in segments(in_ring):
                line_lengths.append(segment.length)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return (
        statistics.mean(line_lengths),
        statistics.stdev(line_lengths) if len(line_lengths) > 1 else 0,
    )


def mean_std_dev_area(geom: BaseGeometry) -> Tuple[float, float]:
    """

    :param geom:
    :return:
    """
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)
    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return (
        statistics.mean(poly_areas),
        statistics.stdev(poly_areas) if len(poly_areas) > 1 else 0,
    )


def prune_area(geom: BaseGeometry, eps: float = 1e-7) -> BaseGeometry:
    """

    :param geom:
    :param eps:
    :return:
    """
    raise NotImplementedError
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)

    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)

    return poly_areas


def prune_rings(geom: BaseGeometry, eps: float = 1e-7) -> BaseGeometry:
    """

    :param geom:
    :param eps:
    :return:
    """
    raise NotImplementedError
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)
    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)

    return poly_areas


def prune_holes(
    geom: Union[MultiPolygon, Polygon], epsilon: Number = 1000
) -> Union[MultiPolygon, Polygon]:
    """

    :param epsilon:
    :param geom:
    :return:"""

    if isinstance(geom, MultiPolygon):
        parts = []

        for polygon in geom.geoms:
            interiors = []

            for interior in polygon.interiors:
                p = Polygon(interior)

                if p.area > epsilon:
                    interiors.append(interior)

            temp_pol = Polygon(polygon.exterior.coords, holes=interiors)
            parts.append(temp_pol)

        return MultiPolygon(parts)

    interiors = []

    for interior in geom.interiors:
        p = Polygon(interior)
        if p.area > epsilon:
            interiors.append(interior)

    return Polygon(geom.exterior.coords, holes=interiors)


def get_coords_from_polygonal_shape(
    shape: Union[shapely.Polygon, shapely.MultiPolygon]
) -> List[List[List[tuple[float, float]]]]:
    coords = []

    if isinstance(shape, shapely.Polygon):
        coords.append(shape.exterior.coords[:])
        for linearring in shape.interiors:
            coords.append(linearring.coords[:])
    elif isinstance(shape, shapely.MultiPolygon):
        for polygon in shape.geoms:
            coords.append(get_coords_from_polygonal_shape(polygon))

    return coords


def get_polygonal_shape_from_coords(
    coords: Union[
        Iterable[Iterable[Iterable[tuple[float, float]]]],
        Iterable[Iterable[tuple[float, float]]],
    ]
) -> Union[Polygon, MultiPolygon, None]:
    outer = next(iter(coords), None)

    assert isinstance(outer, Iterable)
    a = next(iter(outer), None)
    if isinstance(a, Iterable):  # Polygon
        b = next(iter(a), None)
        if isinstance(b, Iterable):  # Holes
            polygons = []
            for poly in coords:  # MultiPolygon and # MultiPolygon Holes
                polygons.append(get_polygonal_shape_from_coords(poly))
            return MultiPolygon(polygons)
        else:
            exterior, *interior = coords
            if interior:
                return Polygon(exterior, holes=interior)
            return Polygon(exterior)

    if len(coords[0]) == 0:
        return None

    return Polygon(coords)


def extract_poly_coords(geom: Union[Polygon, MultiPolygon]) -> Tuple[List, List]:
    """
    TODO: Duplicate of get_coords_from_polygonal_shape

    :param geom:
    :return:
    """
    if geom.type == "Polygon":
        exterior_coords = geom.exterior.coords[:]
        interior_coords = []
        for interior in geom.interiors:
            interior_coords += interior.coords[:]
    elif geom.type == "MultiPolygon":
        exterior_coords = []
        interior_coords = []
        for part in geom.geoms:
            epc = extract_poly_coords(part)  # Recursive call
            exterior_coords += epc[0]
            interior_coords += epc[1]
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")
    return exterior_coords, interior_coords


def extract_poly_rings(geom: BaseGeometry) -> Tuple[List, List]:
    """

    :param geom:
    :return:
    """
    interior_rings = []
    exterior_rings = []
    if isinstance(geom, Polygon):
        exterior_rings.append(geom.exterior)
        interior_rings.extend(geom.interiors)
    elif isinstance(geom, MultiPolygon):
        for part in geom.geoms:
            exterior_rings.append(part.exterior)
            interior_rings.extend(part.interiors)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return exterior_rings, interior_rings


def discard_holes(
    shape: Union[shapely.Polygon, shapely.MultiPolygon]
) -> Union[Polygon, MultiPolygon]:
    if isinstance(shape, shapely.Polygon):
        return shapely.Polygon(shape.exterior.coords)
    elif isinstance(shape, shapely.MultiPolygon):
        shape_parts = []
        for shape_part in shape.geoms:
            shape_parts.append(shapely.Polygon(shape_part.exterior.coords))
        return MultiPolygon(shape_parts)
    raise NotImplementedError


def ensure_ccw_poly(polygon: Polygon) -> Polygon:
    """
    This function checks if the polygon is counter-clockwise if not it is reversed


    :param polygon: The polygon to check
    :return: Returns the polygon turned clockwise
    """

    return Polygon(
        shell=ensure_ccw_ring(polygon.exterior),
        holes=[ensure_ccw_ring(hole) for hole in polygon.interiors],
    )


def ensure_cw_poly(polygon: Polygon) -> Polygon:
    """
    This function checks if the polygon is clockwise if not it is reversed


    :param polygon: The polygon to check
    :return: Returns the polygon turned clockwise
    """

    return Polygon(
        shell=ensure_cw_ring(polygon.exterior),
        holes=[ensure_cw_ring(hole) for hole in polygon.interiors],
    )


def iter_polygons(
    _input_geometry: BaseGeometry,
) -> Union[Generator[Polygon, None, None], Tuple[BaseGeometry]]:
    """

    :param _input_geometry:
    :return:
    """
    if isinstance(_input_geometry, MultiPolygon):
        return (polygon for polygon in _input_geometry.geoms)

    # assert isinstance(_input_geometry, Polygon)

    return (_input_geometry,)


def explode_polygons(
    polygons: Union[Polygon, MultiPolygon, Sequence[Polygon]],
    return_index: bool = False,
) -> Union[Sequence[LineString], Tuple[Sequence[LineString], Sequence[int]]]:
    """

    :param polygons:
    :param return_index:
    :return: main line features that make up the polygons
    """
    lines_out = []
    index = []

    if isinstance(polygons, Polygon):
        polygons = [polygons]

    if isinstance(polygons, MultiPolygon):
        polygons = polygons.geoms

    for i, l in enumerate(polygons):
        for s in [LineString(s) for s in pairs(l.exterior.coords)]:
            lines_out.append(s)
            index.append(i)

        for p in l.interiors:
            lines_out.append(LineString(p.coords))
            index.append(i)

    if return_index:
        return lines_out, index

    return lines_out


if __name__ == "__main__":

    def aishdjauisd():
        # Import constructors for creating geometry collections
        from shapely.geometry import MultiPoint, MultiLineString

        # Import necessary geometric objects from shapely module
        from shapely.geometry import Point, LineString, Polygon

        # Create Point geometric object(s) with coordinates
        point1 = Point(2.2, 4.2)
        point2 = Point(7.2, -25.1)
        point3 = Point(9.26, -2.456)
        # point3D = Point(9.26, -2.456, 0.57)

        # Create a MultiPoint object of our points 1,2 and 3
        multi_point = MultiPoint([point1, point2, point3])

        # It is also possible to pass coordinate tuples inside
        multi_point2 = MultiPoint([(2.2, 4.2), (7.2, -25.1), (9.26, -2.456)])

        # We can also create a MultiLineString with two lines
        line1 = LineString([point1, point2])
        line2 = LineString([point2, point3])
        multi_line = MultiLineString([line1, line2])
        polygon = Polygon([point2, point1, point3])

        from shapely.geometry import GeometryCollection
        from matplotlib import pyplot
        import geopandas

        geoms = GeometryCollection([multi_point, multi_point2, multi_line, polygon])
        print(mean_std_dev_line_length(geoms))
        geoms = sanitise(geoms)
        print(mean_std_dev_line_length(geoms))

        p = geopandas.GeoSeries(geoms)
        p.plot()
        pyplot.show()

    def ahsudh():
        # Import constructors for creating geometry collections
        from shapely.geometry import MultiPoint, MultiLineString

        # Import necessary geometric objects from shapely module
        from shapely.geometry import Point, LineString, Polygon

        # Create Point geometric object(s) with coordinates
        point1 = Point(2.2, 4.2)
        point2 = Point(7.2, -25.1)
        point3 = Point(9.26, -2.456)
        # point3D = Point(9.26, -2.456, 0.57)

        # Create a MultiPoint object of our points 1,2 and 3
        multi_point = MultiPoint([point1, point2, point3])

        # It is also possible to pass coordinate tuples inside
        multi_point2 = MultiPoint([(2.2, 4.2), (7.2, -25.1), (9.26, -2.456)])

        # We can also create a MultiLineString with two lines
        line1 = LineString([point1, point2])
        line2 = LineString([point2, point3])
        multi_line = MultiLineString([line1, line2])
        polygon = Polygon([point2, point1, point3])
        polygon2 = Polygon([point3, point2, point1])
        multi_polygon = MultiPolygon([polygon, polygon2])

        print(mean_std_dev_line_length(line1))
        print(mean_std_dev_line_length(line2))
        print(mean_std_dev_line_length(multi_line))
        print(mean_std_dev_line_length(polygon))
        print(mean_std_dev_line_length(polygon2))
        print(mean_std_dev_line_length(multi_polygon))

        print(mean_std_dev_area(polygon))
        print(mean_std_dev_area(polygon2))
        print(mean_std_dev_area(multi_polygon))

        from shapely.geometry import GeometryCollection
        from matplotlib import pyplot
        import geopandas

        geoms = GeometryCollection(
            [multi_point, multi_point2, multi_line, polygon, polygon2, multi_polygon]
        )

        p = geopandas.GeoSeries(geoms)
        p.plot()
        pyplot.show()

    ahsudh()
    # aishdjauisd()



from typing import Optional, Any

# noinspection PyUnresolvedReferences
from qgis.core import QgsLayerTreeGroup

__all__ = ["duplicate_groups", "select_layer_in_group", "is_group_selected"]


def duplicate_groups(
    group_to_duplicate, *, group_parent: Optional[Any] = None, appendix: str = " (Copy)"
) -> QgsLayerTreeGroup:
    assert (
        group_to_duplicate is not None
    ), f"{group_to_duplicate=} is required to create a duplicate group"

    if group_parent is None:
        group_parent = group_to_duplicate.parent()

    if group_parent is None:
        raise ValueError(f"Group parent was not found for {group_to_duplicate}")

    new_group_parent = group_parent.addGroup(f"{group_to_duplicate.name()}{appendix}")
    for original_group_child in group_to_duplicate.children():
        if isinstance(original_group_child, QgsLayerTreeGroup):
            new_child_group = duplicate_groups(
                original_group_child, group_parent=new_group_parent, appendix=""
            )  # Only top level
        else:
            new_group_parent.addChildNode(original_group_child.clone())

    return new_group_parent


def select_layer_in_group(layer_name, group_name):
    # noinspection PyUnresolvedReferences
    from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer, QgsProject

    # noinspection PyUnresolvedReferences
    from qgis.utils import iface

    group = QgsProject.instance().layerTreeRoot().findGroup(group_name)
    if group is not None:
        for child in group.children():
            if child.name() == layer_name:
                iface.setActiveLayer(child.layer())


def is_group_selected(group_name):
    # noinspection PyUnresolvedReferences
    from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer, QgsProject

    # noinspection PyUnresolvedReferences
    from qgis.utils import iface

    group = QgsProject.instance().layerTreeRoot().findGroup(group_name)
    return group in iface.layerTreeView().selectedNodes()



#!/usr/bin/env python3

from functools import partial
from typing import Callable

from jord.qlive_utilities import (
    QliveClient,
    QliveRPCMethodEnum,
    QliveRPCMethodMap,
    build_package,
)

__doc__ = r"""Warning this client a no client-side validation, which might result in server-side exceptions"""

__all__ = ["AutoQliveClient"]

from jord.qlive_utilities.clients.arguments import partial_satisfied


class AutoQliveClient(QliveClient):
    """
    Has no client side validation of data, but exposes function if available in QliveRPCMethodEnum and QliveRPCMethodMap
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for method in QliveRPCMethodEnum:
            if method in QliveRPCMethodMap:
                actual_callable = QliveRPCMethodMap[method]

                if False:
                    partial_build_package = partial(build_package, method)
                    if False and partial_satisfied(
                        partial_build_package
                    ):  # TODO: RESOLVE PARTIAL APPLICATION SATISFACTION.

                        def a():
                            self.send(partial_build_package())

                        rpc_method = a
                    elif True:

                        def a(*args):
                            self.send(partial_build_package(*args))

                        rpc_method = a
                    elif False:

                        def a(**kwargs):
                            self.send(partial_build_package(**kwargs))

                        rpc_method = a
                    elif False:

                        def a(*args, **kwargs):
                            self.send(partial_build_package(*args, **kwargs))

                        rpc_method = a
                    else:
                        raise NotImplementedError
                elif False:
                    rpc_method = lambda *args: self.send(
                        partial(build_package, method)(*args)
                    )
                elif False:
                    rpc_method = lambda *args: self.send(build_package(method, *args))
                else:

                    def wrapped(method_, *args_, **_kwargs) -> Callable:
                        return self.send(build_package(method_, *args_, **_kwargs))

                    rpc_method = partial(wrapped, method)

                rpc_method.__doc__ = actual_callable.__doc__
                setattr(self, method.value, rpc_method)


if __name__ == "__main__":
    # QliveClient().clear_all()
    # QliveClient().remove_layers()
    # print(QliveClient().clear_all.__doc__)
    # print(QliveClient().__dict__)
    def uahdsuh():
        with AutoQliveClient() as qlive:
            qlive.add_wkts({"a": "POINT (-66.86 10.48)"})

    # QliveClient().add_dataframe(None)

    uahdsuh()
